/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                                                                        **
**                           BIALYSTOK 2007                               **
**                                                                        **
**                             MODULE IP                                  **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2007.03.22                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/


#include <LPC23xx.H>
#include "ipservice.h"
#include "system.h"
#include "types.h"
#include "ipmemservice.h"
#include "frameutil.h"
#include "arpservice.h"
#include "icmpservice.h"
#include "systime.h"
#include "irqvectors.h"
#include "uart0.h"
#include "trace.h"

#define StaticQueueSize 8

typedef struct {
                 USHORT                  WriteIndex ;
                 USHORT                  ReadIndex ;
                 PtrOSFrameRecT          FrameBuffer [ StaticQueueSize ] ;
//                 PtrInpFramesQueueRecT   DynamicQueue ;
               } CyclicFrameQueueRecT ;

#define SendQMaxLifeTime      24

typedef struct {
                 ULONG            DestIPAddress ;
                 USHORT           LifeTime ;
                 PtrOSFrameRecT   Frame ;
               } SendQueueElementRecT ;

#define SendQueueSize        16
static SendQueueElementRecT SendQueue [ SendQueueSize ] ;
static IPLayerInstanceRecT LocalIPLayerInstance ;
static CyclicFrameQueueRecT CyclicQueue ;
static volatile USHORT IPTimerCallCounter ;
static volatile USHORT TimeEventSignal ;

//static void AddToDynamicQueue ( PtrOSFrameRecT Frame )
//{
//  PtrInpFramesQueueRecT NewQueueElement ;
//  /*-------------------------------------------------------------------------*/
//  NewQueueElement = ( PtrInpFramesQueueRecT ) malloc ( sizeof ( InpFramesQueueRecT ) ) ;
//  NewQueueElement -> Frame = Frame ;
//  NewQueueElement -> FwdLink = IPLayerInstance . CyclicFrameQueue . DynamicQueue ;
//  IPLayerInstance . CyclicFrameQueue . DynamicQueue = NewQueueElement ;
//} /* AddToDynamicQueue */


static void AddFrameToQueue ( PtrOSFrameRecT Frame )
{
  USHORT NextIndex ;
  /*-------------------------------------------------------------------------*/
  NextIndex = ( CyclicQueue . WriteIndex + 1 ) % StaticQueueSize ;
  if ( NextIndex == CyclicQueue . ReadIndex )
  {
//      AddToDynamicQueue ( Frame ) ;
  } /* if ... */
  else
  {
    CyclicQueue . FrameBuffer [ CyclicQueue . WriteIndex ] = Frame ;
    CyclicQueue . WriteIndex = NextIndex ;
  } /* if ... else */ ;
} /* AddFrameToQueue  */


static PtrOSFrameRecT ExcludeFromQueue ( void )
{
  PtrOSFrameRecT Frame ;
  /*-------------------------------------------------------------------------*/
  if ( CyclicQueue . ReadIndex != CyclicQueue . WriteIndex )
  {
    Frame = CyclicQueue . FrameBuffer [ CyclicQueue . ReadIndex ] ;
    CyclicQueue . ReadIndex = ( CyclicQueue . ReadIndex + 1 ) % StaticQueueSize ;
    return ( Frame ) ;
  } /* if */ ;
  return NIL ;
} /* ExcludeFromQueue */

UCHAR IPMessage10 [ ]          = "\r\nNaglowek IP:" ;
UCHAR IPMessage11 [ ]          = "\r\nVERS=" ;
UCHAR IPMessage12 [ ]          = " HLEN=" ;
UCHAR IPMessage13 [ ]          = " TOS=" ;
UCHAR IPMessage14 [ ]          = " TOTAL=" ;
UCHAR IPMessage15 [ ]          = "\r\nIDENT=" ;
UCHAR IPMessage16 [ ]          = " FLAGS=" ;
UCHAR IPMessage17 [ ]          = " FRAGOFFS=" ;
UCHAR IPMessage18 [ ]          = "\r\nTTL=" ;
UCHAR IPMessage19 [ ]          = " PROT=" ;
UCHAR IPMessage20 [ ]          = " HCRC=" ;
UCHAR IPMessage21 [ ]          = "\r\nSRC=" ;
UCHAR IPMessage22 [ ]          = "\r\nDST=" ;
UCHAR IPMessage23 [ ]          = "\r\nJest to pakiet ICMP\r\n\r\n" ;
UCHAR IPMessage24 [ ]          = "\r\nJest to pakiet TCP\r\n\r\n" ;
UCHAR IPMessage25 [ ]          = "\r\nJest to pakiet UDP\r\n\r\n" ;

static void ProcessIPPacked ( PtrIPLayerInstanceRecT IPLayerInstance ,
                              PtrOSFrameRecT         Frame )
{
  IPHeaderRecT IPHeader ;
  /*-------------------------------------------------------------------------*/
  GetIPHeader ( & IPHeader , Frame ) ;
  UART0SendString ( IPMessage10 ) ;
  UART0SendString ( IPMessage11 ) ;
  TraceUCHAR ( IPHeader . Vers ) ;
  UART0SendString ( IPMessage12 ) ;
  TraceUCHAR ( IPHeader . HLen ) ;
  UART0SendString ( IPMessage13 ) ;
  TraceUCHAR ( IPHeader . ServiceType ) ;
  UART0SendString ( IPMessage14 ) ;
  TraceUSHORT ( IPHeader . TotalLength ) ;
  UART0SendString ( IPMessage15 ) ;
  TraceUSHORT ( IPHeader . Identification ) ;
  UART0SendString ( IPMessage16 ) ;
  TraceUCHAR ( IPHeader . Flags ) ;
  UART0SendString ( IPMessage17 ) ;
  TraceUSHORT ( IPHeader . FragmentOffs ) ;
  UART0SendString ( IPMessage18 ) ;
  TraceUCHAR ( IPHeader . TTL ) ;
  UART0SendString ( IPMessage19 ) ;
  TraceUSHORT ( IPHeader . Protocol ) ;
  UART0SendString ( IPMessage20 ) ;
  TraceUSHORT ( IPHeader . HeaderChecksum ) ;
  UART0SendString ( IPMessage21 ) ;
  TraceIP ( IPHeader . SourceIPAddress ) ;
  UART0SendString ( IPMessage22 ) ;
  TraceIP ( IPHeader . DestinationIPAddress ) ;
  if ( IPHeader . Vers == IPV4 )
  {
    if ( IPHeader . HLen == 5 )
    {
      switch ( IPHeader . Protocol )
      {
        case Prot_ICMP                :
          UART0SendString ( IPMessage23 ) ;
          if ( IPHeader . DestinationIPAddress == IPLayerInstance -> MyLocalIPAddress )
          {
            IPLayerInstance -> ICMPProcessService ( IPLayerInstance , & IPHeader , Frame ) ;
          } /* if */ ;
          break ;
        case Prot_TCP                 :
          UART0SendString ( IPMessage24 ) ;
          if ( IPHeader . DestinationIPAddress == IPLayerInstance -> MyLocalIPAddress )
          {
            if ( IPLayerInstance -> TCPProcessService )
              IPLayerInstance -> TCPProcessService ( IPLayerInstance , & IPHeader , Frame ) ;
          } /* if */ ;
          break ;
        case Prot_UDP                 :
          UART0SendString ( IPMessage25 ) ;
          if ( ( IPHeader . DestinationIPAddress == IPLayerInstance -> MyLocalIPAddress ) ||
               ( ( IPHeader . DestinationIPAddress & ( ~ IPLayerInstance -> MyLocalNetMask ) ) == ( ~ IPLayerInstance -> MyLocalNetMask ) ) )
          {
            if ( IPLayerInstance -> UDPProcessService )
              IPLayerInstance -> UDPProcessService ( IPLayerInstance , & IPHeader , Frame ) ;
          } /* if */ ;
          break ;
        default :
          ;
      } /* switch */ ;
    } /* if */ ;
  } /* if */ ;
} /* ProcessIPPacked */


static void ProcessEthFrame ( PtrIPLayerInstanceRecT IPLayerInstance ,
                              PtrOSFrameRecT         Frame )
{
  /*-------------------------------------------------------------------------*/
} /* ProcessEthFrame */


static void CopyFrame (  PtrOSFrameRecT OutFrame ,
                         PtrOSFrameRecT InFrame )
{
  ULONG Loop ;
  UCHAR * OutBuffer ;
  UCHAR * InBuffer ;
  /*-------------------------------------------------------------------------*/
  OutFrame -> Length = InFrame -> Length ;
  OutFrame -> Index = InFrame -> Index ;
  OutBuffer = OutFrame -> Data ;
  InBuffer = InFrame -> Data ;
  for ( Loop = 0 ; Loop < InFrame -> Length ; Loop ++ )
    * OutBuffer ++ = * InBuffer ++ ;
} /* CopyFrame */


static void AddPackedToSendQueue ( PtrIPLayerInstanceRecT IPLayerInstance ,
                                   PtrOSFrameRecT         Frame ,
                                   ULONG                  IPAddress )
{
  EMACInstanceRecT * InterfInstance ;
  PtrOSFrameRecT OutFrame ;
  ULONG Loop ;
  /*-------------------------------------------------------------------------*/
  InterfInstance = GetEMACInstance ( IPLayerInstance ) ;
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  CopyFrame ( OutFrame , Frame ) ;
  for ( Loop = 0 ; Loop < SendQueueSize ; Loop ++ )
  {
    if ( ! SendQueue [ Loop ] . Frame )
    {
      SendQueue [ Loop ] . Frame = OutFrame ;
      SendQueue [ Loop ] . LifeTime = SendQMaxLifeTime ;
      SendQueue [ Loop ] . DestIPAddress = IPAddress ;
      return ;
    } /* IF */ ;
  } /* for */ ;
} /* AddPackedToSendQueue */


static void SendIPPacked ( PtrIPLayerInstanceRecT IPLayerInstance ,
                           PtrOSFrameRecT         Frame ,
                           ULONG                  IPAddress )
{
  PtrARPTableElementRecT ARPTableElement ;
  EMACInstanceRecT * InterfInstance ;
  ULONG TargetAddress ;
  /*-------------------------------------------------------------------------*/
  InterfInstance = GetEMACInstance ( IPLayerInstance ) ;
  if ( ( IPAddress & IPLayerInstance -> MyLocalNetMask ) ==
       ( IPLayerInstance -> MyLocalIPAddress & IPLayerInstance -> MyLocalNetMask ) )
  {
    TargetAddress = IPAddress ;
  } /* if ... */
  else
  {
    TargetAddress = IPLayerInstance-> MyDefaultGWIPAddress ;
    if ( ! TargetAddress )
    {
      return ;
    } /* if */ ;
  } /* if ... else */ ;
  ARPTableElement = LocateMACForIP ( IPLayerInstance , TargetAddress ) ;
  if ( ! ARPTableElement )
  {
    PrepareARPRequest ( IPLayerInstance , TargetAddress ) ;
    IPLayerInstance -> AddToSendQueueService ( IPLayerInstance , Frame , TargetAddress ) ;
  } /* if ... */
  else
  {
    PutByteArrayField ( Frame , 0 , ARPTableElement -> MACAddress , EthMACAddressSize ) ;
    InterfInstance -> InterfaceSendFrame ( Frame ) ;
  } /* if ... else */ ;
} /* SendIPPacked */

UCHAR IPMessage01 [ ]              = "\r\nOdebrano ramke do MAC=" ;
UCHAR IPMessage02 [ ]              = " od MAC=" ;
UCHAR IPMessage03 [ ]              = "Jest to pakiet ARP skierowany do IP=" ;
UCHAR IPMessage04 [ ]              = "Jest to pakiet IP\r\n" ;

static void IPLayerPool ( PtrIPLayerInstanceRecT IPLayerInstance )
{
  PtrOSFrameRecT Frame ;
  EtherPackedHeaderRecT EtherPackedHeader ;
  EMACInstanceRecT * InterfInstance ;
  PtrARPTableElementRecT ARPTableElement ;
  ULONG Loop ;
  /*-------------------------------------------------------------------------*/
  InterfInstance = GetEMACInstance ( IPLayerInstance ) ;
  InterfInstance -> CheckEthInterface ( ) ;
  Frame = ExcludeFromQueue ( ) ;
  if ( Frame )
  {
    GetEthHeader ( & EtherPackedHeader , Frame ) ;
    UART0SendString ( IPMessage01 ) ;
    TraceMAC ( EtherPackedHeader . DestinMACAddress ) ;
    UART0SendString ( IPMessage02 ) ;
    TraceMAC ( EtherPackedHeader . SourceMACAddress ) ;
    UART0Send ( 0x0D ) ;
    UART0Send ( 0x0A ) ;
    SetIPHeaderOrgin ( Frame ) ; 
    if ( EtherPackedHeader . LengthOrType >= 1518 )
    {
      switch ( EtherPackedHeader . LengthOrType )
      {
        case FRAME_ARP    :
          UART0SendString ( IPMessage03 ) ;
          IPLayerInstance -> ARPProcessService ( IPLayerInstance , Frame ) ;
          break ;
        case FRAME_IP     :
          UART0SendString ( IPMessage04 ) ;
          IPLayerInstance -> IPProcessService ( IPLayerInstance , Frame ) ;
          break ;
        default :
          IPLayerInstance -> OthFrameProcessSerevice ( IPLayerInstance , Frame ) ;
      } /* switch */ ;
    } /* if ... */
    else
    {
      IPLayerInstance -> OthFrameProcessSerevice ( IPLayerInstance , Frame ) ;
    } /* if ... else */ ;
    InterfInstance -> InterfaceDeallocFrameMemory ( Frame ) ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < SendQueueSize ; Loop ++ )
  {
    Frame = SendQueue [ Loop ] . Frame ;
    if ( Frame )
    {
      ARPTableElement = LocateMACForIP ( IPLayerInstance , SendQueue [ Loop ] . DestIPAddress ) ;
      if ( ARPTableElement )
      {
        PutByteArrayField ( Frame , 0 , ARPTableElement -> MACAddress , EthMACAddressSize ) ;
        InterfInstance -> InterfaceSendFrame ( Frame ) ;
        SendQueue [ Loop ] . Frame = NIL ;
        InterfInstance -> InterfaceDeallocFrameMemory ( Frame ) ;
        break ;
      } /* if */ ;
    } /* if */ ;
  } /* for */ ;
  if ( TimeEventSignal )
  {
    TimeEventSignal = 0 ;
    IPLayerInstance -> StandardTimeService ( IPLayerInstance ) ;
    if ( IPLayerInstance -> ExtraTimeService )
      IPLayerInstance -> ExtraTimeService ( IPLayerInstance ) ;
  } /* if */ ;
} /* IPLayerPool */


static void IPTimeService ( PtrIPLayerInstanceRecT IPLayerInstance )
{
  ULONG Loop ;
  PtrOSFrameRecT Frame ;
  EMACInstanceRecT * InterfInstance ;
  /*-------------------------------------------------------------------------*/
  IPLayerInstance -> DiagARPTableService ( IPLayerInstance ) ;
  InterfInstance = GetEMACInstance ( IPLayerInstance ) ;
  for ( Loop = 0 ; Loop < SendQueueSize ; Loop ++ )
  {
    Frame = SendQueue [ Loop ] . Frame ;
    if ( Frame )
    {
      if ( ! ( SendQueue [ Loop ] . LifeTime -- ) )
      {
        SendQueue [ Loop ] . Frame = NIL ;
        InterfInstance -> InterfaceDeallocFrameMemory ( Frame ) ;
      } /* if */ ;
    } /* if */ ;
  } /* for */ ;
  if ( IPLayerInstance -> TCPTimerPoolService )
    IPLayerInstance -> TCPTimerPoolService ( IPLayerInstance -> TCPInstanceEntry ) ;
} /* IPTimeService */ ;


__irq void IPTimerIQRHandler ( void )
{
  /*-------------------------------------------------------------------------*/
  SysTimeEnevtService ( ) ;
  IPTimerCallCounter ++ ;
  if ( IPTimerCallCounter >= 100 )
  {
    IPTimerCallCounter = 0 ;
    TimeEventSignal = 1 ;
  } /* if */ ;
  T0IR = 1 ;
  VICVectAddr = 0 ;
} /* IPTimerIQRHandler */


static void InitTimerService ( void )
{
  /*-------------------------------------------------------------------------*/
  IPTimerCallCounter = 0 ;
  TimeEventSignal = 0 ;
  T0MR0 = ( PCLK / 1000 ) - 1 ;
  /*14399 ;*/
  T0MCR = 3 ;
  T0TCR = 1 ;
  VICVectAddr4 = ( unsigned long ) IPTimerIQRHandler ;
  VICVectCntl4 = 15 ;
  VICIntEnable = ( 1 << TIMER0_Interrupt ) ;
} /* InitTimerService */


void SetIPAdresses ( PtrIPLayerInstanceRecT IPLayerInstance ,
                     ULONG                  IPAddress ,
                     ULONG                  NetMask ,
                     ULONG                  GateWayIP )
{
  /*-------------------------------------------------------------------------*/
  IPLayerInstance -> MyLocalIPAddress = IPAddress ;
  IPLayerInstance -> MyLocalNetMask = NetMask ;
  IPLayerInstance -> MyDefaultGWIPAddress = GateWayIP ;
} /* SetIPAdresses */


PtrEMACInstanceRecT GetEMACInstance ( PtrIPLayerInstanceRecT IPLayerInstance )
{
  /*-------------------------------------------------------------------------*/
  return ( IPLayerInstance -> InterfaceInstance ) ;
} /* GetEMACInstance */


PtrIPLayerInstanceRecT InitIPLayerInstance ( EMACInstanceRecT * EMACInstance ,
                                             ULONG              IPAddress ,
                                             ULONG              NetMask )
{
  USHORT Loop ;
  /*-------------------------------------------------------------------------*/
  InitTimerService ( ) ;
  InitIPLayerMemoryService ( ) ;
  for ( Loop = 0 ; Loop < SendQueueSize ; Loop ++ )
    SendQueue [ Loop ] . Frame = NIL ;
  for ( Loop = 0 ; Loop < ARPTableSize ; Loop ++ )
    LocalIPLayerInstance . ARPTable [ Loop ] . ValidElement = 0 ;
  LocalIPLayerInstance . MyLocalIPAddress = IPAddress ;
  LocalIPLayerInstance . MyLocalNetMask = NetMask ;
  LocalIPLayerInstance . MyDefaultGWIPAddress = 0 ;
  AutoIdentyficationCounter = 0x0001 ;
  LocalIPLayerInstance . UDPInstanceEntry = NIL ;
  LocalIPLayerInstance . TCPInstanceEntry = NIL ;
  LocalIPLayerInstance . StandardTimeService = IPTimeService ;
  LocalIPLayerInstance . ExtraTimeService = NIL ;
  LocalIPLayerInstance . DiagARPTableService = DiagARPTable ;
  LocalIPLayerInstance . InterfaceInstance = EMACInstance ;
  LocalIPLayerInstance . OthFrameProcessSerevice = ProcessEthFrame ;
  LocalIPLayerInstance . ARPProcessService = ProcessARPPacked ;
  LocalIPLayerInstance . ICMPProcessService = ProcessICMPPacked ;
  LocalIPLayerInstance . ICMPEchoProcessService = NIL ;
  LocalIPLayerInstance . UDPProcessService = NIL ;
  LocalIPLayerInstance . TCPProcessService = NIL ;
  LocalIPLayerInstance . TCPTimerPoolService = NIL ;
  LocalIPLayerInstance . IPProcessService = ProcessIPPacked ;
  LocalIPLayerInstance . IPLayerPoolService = IPLayerPool ;
  LocalIPLayerInstance . SendIPPackedService = SendIPPacked ;
  LocalIPLayerInstance . AddToSendQueueService = AddPackedToSendQueue ;
  EMACInstance -> InterfaceProcessFrame = AddFrameToQueue ;
  EMACInstance -> InterfaceAllocFrameMemory = FrameMemoryAllocate ;
  EMACInstance -> InterfaceDeallocFrameMemory = FrameMemoryDeallocate ;
  return ( & LocalIPLayerInstance ) ;
} /* InitIPLayerInstance */


